from typing import Tuple, List, Dict


class Disambiguator:
    """ Function or class that takes a list of words and returns the disambiguated list of words as tuples"""
    def __call__(self, sent, tasks) -> Tuple[List[str], List[Dict[str, str]]]:
        raise NotImplementedError



import regex as re
from .proto import Formatter


class GlueFormatter(Formatter):
    """ Need replacing of morph_part for specific corpora

    """

    HEADERS = ["form", "lemma", "POS", "morph", "treated_token"]
    MORPH_PART = ["Case", "Numb", "Deg", "Mood", "Tense", "Voice", "Person"]
    PONCTU = re.compile(r"^\W+$")

    def __init__(self, tokenizer_memory):
        super(GlueFormatter, self).__init__([])
        self.tokenizer_memorytokenizer_memory = tokenizer_memory

    def __call__(self, tasks):
        super(GlueFormatter, self).__init__(tasks)
        self.pos_tag = "POS"
        if "POS" not in self.tasks and "pos" in self.tasks:
            self.pos_tag = "pos"
        return self

    @classmethod
    def get_headers(cls):
        return cls.HEADERS

    def rule_based(cls, token):
        if cls.PONCTU.match(token):
            return [token, token, "PUNC", "MORPH=empty", token]

        return None

    def format_line(self, token, tags, ignored=False):
        tags = list(tags)
        lemma = tags[self.tasks.index("lemma")]
        index, input_token, out_token = self.tokenizer_memory.tokens.pop(0)
        if token != out_token:
            raise Exception("The output token does not match our inputs %s : %s" % (token, out_token))

        overwriten = self.rule_based(token)
        if overwriten:
            return overwriten

        return [
            input_token,
            lemma,
            tags[self.tasks.index(self.pos_tag)],
            "|".join(
                "{cat}={tag}".format(
                    cat=morph_part,
                    tag=tags[self.tasks.index(morph_part)]
                )
                for morph_part in type(self).MORPH_PART
                if morph_part in self.tasks and
                tags[self.tasks.index(morph_part)] != "_"
            ) or "MORPH=empty",
            out_token
        ]



import re
import string

from pie.tagger import simple_tokenizer
from typing import Callable, List, Tuple, Dict, Union, Iterable

from ...pipeline.tokenizers.classes import Tokenizer
from ...utils import ObjectCreator

Remover = Callable[[List[str]], Tuple[List[str], Dict[int, str]]]
PUNKT = re.compile("^["+string.punctuation+"]+$")


class DataIterator:
    def __init__(self, tokenizer: Union[ObjectCreator, Tokenizer] = None, remove_from_input: Callable = None):
        """ Iterator used to parse the text and returns bits to tag

        :param tokenizer: Tokenizer
        """
        self.tokenizer = tokenizer or simple_tokenizer
        self.remove_from_input = remove_from_input
        if self.remove_from_input is None:
            self.remove_from_input = lambda x: (x, {})

    @staticmethod
    def remove_punctuation(sentence: List[str]) -> Tuple[List[str], Dict[int, str]]:
        """ Removes punctuation from a list and keeps its index

        :param sentence:
        :return: First the sentence with things removed, then a dictionary whose keys are index of token to reinsert and
        associated values are punctuation to reinsert.

        >>> x = DataIterator.remove_punctuation(["Je", "suis", "content",",", "mais", "...", '"', "fatigué", '"', "."])
        >>> assert x == (['Je', 'suis', 'content', 'mais', 'fatigué'], {3: ',', 5: '...', 6: '"', 8: '"', 9: '.'})
        """
        clean, removed = [], {}
        for index, token in enumerate(sentence):
            if PUNKT.match(token):
                removed[index] = token
            else:
                clean.append(token)
        return clean, removed

    def get_tokenizer(self) -> Tokenizer:
        """ Get the tokenizer if it needs to be created"""
        if isinstance(self.tokenizer, ObjectCreator):
            return self.tokenizer.create()
        return self.tokenizer

    def get_remover(self) -> Remover:
        if isinstance(self.remove_from_input, ObjectCreator):
            return self.remove_from_input.create()
        return self.remove_from_input

    def __call__(self, data: str, lower: bool = False) -> Iterable[Tuple[List[str], int, Dict[int, str]]]:
        """ Default iter data takes a text, an option to make lower
        and yield lists of words along with the length of the list

        :param data: A plain text
        :param lower: Whether or not to lower the text
        :yields: (Sentence as a list of word, Size of the sentence, Elements removed from the sentence)
        """
        tokenizer = self.get_tokenizer()
        remover = self.get_remover()
        for sentence in tokenizer(data, lower=lower):
            clean_sentence, removed_from_input = remover(sentence)
            yield clean_sentence, len(clean_sentence), removed_from_input



from collections import namedtuple
import os


Metadata = namedtuple("Metadata", ["title", "lang", "authors", "description", "link"])

_File = namedtuple("File", ["url", "name"])


class File(_File):
    def __str__(self):
        return self.name


PATH = os.path.normpath(
    os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "downloads"
    )
)


def get_path(module, file):
    return os.path.join(PATH, module, file)


class ObjectCreator:
    """ Some objects should be reset everytime a new tagging is done. To make this easier
    we provide this class that keeps in memory the initialization parameters."""
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def create(self):
        return self.cls(*self.args, **self.kwargs)



import click

from . import sub


MODELS = [name for name, *_ in sub.get_list()]


@click.group("pie-ext")
def pie_ext():
    """ Addon for Pie that allows you to download and install modules that you might be interested in """


@pie_ext.command("list")
@click.option("--full", is_flag=True, help="Display full information on models")
def print_list(full):
    """ List available model to download and use """
    click.echo("Available Models:")
    for module, desc in sub.get_list():
        click.echo("\t- {} ({}) : {}".format(
            click.style(module, bold=True),
            desc.lang,
            click.style(desc.title, underline=True)
        ))
        if full:
            click.echo("\t\tAuthor(s): {}".format(", ".join(desc.authors)))
            click.echo("\t\tDescription(s): {}".format(desc.description))
            click.echo("\t\tURL: {}".format(desc.link))


@pie_ext.command("download")
@click.argument("model", type=click.Choice(MODELS, case_sensitive=False))
def download(model):
    """ Download a [model] for future availability. Check list for available models"""
    click.echo(click.style("Starting downloading...", bold=True))
    for index, file in enumerate(sub.download(model)):
        if index == 0:
            click.echo("{} files to download".format(file))
        else:
            click.echo("- {} downloaded".format(file))
    click.echo(click.style("Finished !", bold=True))


@pie_ext.command("tag")
@click.argument("model", type=click.Choice(MODELS, case_sensitive=False))
@click.argument("filepath", nargs=-1, type=click.Path(exists=True, dir_okay=False))
@click.option("--allow-n-failures", "allowed_failure", type=int, default=5)
@click.option("--batch_size", type=int, default=16)
@click.option("--device", type=str, default="cpu")
@click.option("--debug", is_flag=True)
def tag(model, filepath, allowed_failure, batch_size, device, debug):
    """ Tag as many [filepath] as you want with [model]"""
    from tqdm import tqdm
    click.echo(click.style("Getting the tagger", bold=True))
    tagger = sub.get_tagger(model, batch_size=batch_size, device=device)
    failures = []
    for file in tqdm(filepath):
        try:
            sub.tag_file(model, tagger, file)
        except Exception as E:
            failures.append(E)
            click.echo("{} could not be lemmatized".format(file))
            if debug:
                raise E
            if len(failures) > allowed_failure:
                click.echo(
                    click.style("Too many errors, stopping the process", fg="red")
                )
                for failure in failures:
                    print(failure)
                break


@pie_ext.command("install-addons")
@click.argument("model", type=click.Choice(MODELS, case_sensitive=False))
def install(model):
    """ Install addons for specific models"""
    click.echo(click.style("Installing add-ons", bold=True))
    sub.get_addons(model)
    click.echo(click.style("Done", bold=True))



from ...utils import Metadata, File ,get_path
from .classes import GlueFormatter
from ...pipeline.iterators.proto import DataIterator

DESC = Metadata(
    "Ancien Francais (ENC)",
    "fro",
    ["Jean-Baptiste Camps", "Ariane Pinche", "Thibault Clérice", "Frédéric Duval", "Lucence Ing"],
    "Model trained on diverse corpora (hagiography, geste, etc.)",
    "https://github.com/chartes/deucalion-model-af"
)

DOWNLOADS = [
    File("https://github.com/chartes/deucalion-model-af/raw/master/morph.tar", "morph.tar"),
    File("https://github.com/chartes/deucalion-model-af/raw/master/lemma-pos.tar", "lemma-pos.tar")
]


Models = "<{},MODE,TEMPS,PERS,NOMB,GENRE,CAS,DEGRE><{},lemma,pos>".format(
    get_path("fro", "morph.tar"),
    get_path("fro", "lemma-pos.tar")
)


def get_iterator_and_formatter():
    formatter = GlueFormatter
    iterator = DataIterator(
        remove_from_input=DataIterator.remove_punctuation
    )
    return iterator, formatter



from ...utils import Metadata, File, get_path, ObjectCreator
from autocat import NeedsDisambiguation, StraightAutodisambiguation, CategoryAutodisambiguation, GroupAutodisambiguation
from ...pipeline.disambiguators.autocat import Autocat


DESC = Metadata(
    "LASLA-ENC",
    "lat",
    ["Thibault Clérice"],
    "Model trained on LASLA data without disambiguation",
    "https://github.com/chartes/deucalion-model-lasla"
)

DOWNLOADS = [
    File("https://github.com/chartes/deucalion-model-lasla/raw/master/latin-straight.json", "latin-straight.json"),
    File("https://github.com/chartes/deucalion-model-lasla/raw/master/latin-pos.json", "latin-pos.json"),
    File("https://github.com/chartes/deucalion-model-lasla/raw/master/latin-needs.json", "latin-needs.json"),
    File("https://github.com/chartes/deucalion-model-lasla/raw/master/model.tar", "model.tar")
]

Models = "<{},lemma,Voice,Mood,Deg,Numb,Person,Tense,Case,Gend,pos>".format(get_path("lasla", "model.tar"))


def _get_disambiguator():
    pos = CategoryAutodisambiguation.from_file(
        get_path("lasla", "latin-pos.json"), category_key="pos", lemma_key="lemma")
    straight = StraightAutodisambiguation.from_file(
        get_path("lasla", "latin-straight.json"), lemma_key="lemma")
    impossible = NeedsDisambiguation.from_file(
        get_path("lasla", "latin-needs.json"), lemma_key="lemma")
    return Autocat(GroupAutodisambiguation(lemma_key="lemma", categorizers=(straight, pos, impossible)))


Disambiguator: ObjectCreator = ObjectCreator(_get_disambiguator)


def addons():
    from cltk.corpus.utils.importer import CorpusImporter

    corpus_importer = CorpusImporter('latin')
    corpus_importer.import_corpus('latin_models_cltk')



class MemorizingTokenizer(object):
    """ Tokenizer that memoryze what it tokenized.

    Mostly used to normalized input as input time and then reinserting normalized input

    """
    @staticmethod
    def _sentence_tokenizer(string):
        for s in string.split("."):
            if s.strip():
                yield s.strip() + " " + "."

    @staticmethod
    def _word_tokenizer(string):
        for s in string.split():
            if s.strip:
                yield s.strip()

    @staticmethod
    def _replacer(inp: str):
        return inp

    def __init__(self, sentence_tokenizer=None, word_tokenizer=None, replacer=None, normalizer=None):
        self.tokens = [
        ]

        self.re_sentence_tokenizer = sentence_tokenizer or self._sentence_tokenizer
        self.word_tokenizer = word_tokenizer or self._word_tokenizer
        self.replacer = replacer or self._replacer
        self.normalizer = normalizer or self._replacer

    def __call__(self, data, lower=True):
        if lower:
            data = data.lower()

        for sentence in self.sentence_tokenizer(data):
            toks = self.word_tokenizer(sentence)
            new_sentence = []

            for tok in toks:
                if tok:
                    out = self.replacer(tok)
                    self.tokens.append((len(self.tokens), tok, out))
                    new_sentence.append(out)
            if new_sentence:
                yield new_sentence



from typing import List, Iterable


class Formatter:  # Default is TSV
    def __init__(self, tasks: List[str]):
        self.tasks: List[str] = tasks

    def format_line(self, token: str, tags: Iterable[str], ignored=False) -> List[str]:
        """ Format the tags"""
        return [token] + list(tags)

    def write_line(self, formatted):
        return "\t".join(formatted) + "\r\n"

    def write_sentence_beginning(self) -> str:
        return ""

    def write_sentence_end(self) -> str:
        return ""

    def write_footer(self) -> str:
        return ""

    def get_headers(self):
        return ["token"] + self.tasks

    def write_headers(self)-> str:
        """ Format the headers """
        return self.write_line(self.get_headers())



from pie_extended.tagger import ExtensibleTagger


class FakeTagger(ExtensibleTagger):
    def __init__(self, tokens, tasks):
        self.tokens = tokens
        self.tasks = tasks
        self.seen = []
        self.lower = False
        self.batch_size = 100
        self.disambiguation = None

    def tag(self, sents, **kwargs):
        self.seen.extend(sents)

        return self.tokens, self.tasks



from autocat.autocat import ProtoDisambiguator
from .proto import Disambiguator


class Autocat(Disambiguator):
    def __init__(self, autocat_categorizer: ProtoDisambiguator, separator: str = "", **kwargs):
        self.autocat_categorizer: ProtoDisambiguator = autocat_categorizer
        self.lemma_key = autocat_categorizer.lemma_key
        self.separator = separator

    def __call__(self, sent, tasks):
        toks, sents_as_rows = zip(*[
            (token, dict(zip(tasks, tags)))
            for token, tags in sent
        ])

        out = [
        ]

        for (tok, tags, dis) in zip(toks, sents_as_rows, self.autocat_categorizer.disambiguate_rows(sents_as_rows)):
            if dis:
                tags[self.lemma_key] += dis
            out.append((tok, list(tags.values())))

        return out




from pie_extended.models.lasla.classes import get_iterator_and_formatter
from pie_extended.testing_utils import FakeTagger
from typing import List, Tuple

from unittest import TestCase


def make_fake_data(sentences: List[str], nb_tasks: int = 9) -> List[Tuple[str, List[str]]]:
    return [
        [
            (token, tuple([token, *["fake"]*nb_tasks]))
            for token in sentence.split()
        ]
        for sentence in sentences
    ]


def make_controller(sentences: List[str]):
    # Add the lemmatizer routes
    tagger = FakeTagger(
        make_fake_data(sentences),
        tasks="lemma,Voice,Mood,Deg,Numb,Person,Tense,Case,Gend,pos".split(",")
    )
    iterator, formatter = get_iterator_and_formatter()
    return tagger, iterator, formatter


class TestPonctuation(TestCase):
    def test_consecutive_dots(self):
        """Check that consecutive punctation does not break anything

        Found out the hard way it would break things
        """

        tagger, data_iterator, formatter = make_controller([
            "id enim ait turbabuntur a facie eius patris or phanorum et iudicis uiduarum",
            "causam turbationis hanc docuit quod pater"
        ])

        result = tagger.tag_str(
            data="id enim ait turbabuntur a facie eius patris or phanorum et iudicis uiduarum ."
                          "  .  causam turbationis hanc docuit quod pater",
            formatter_class=formatter,
            iterator=data_iterator
        )
        self.assertIn(
            "uiduarum	uiduarum	fake	Case=fake|Numb=fake|Deg=fake|Mood=fake|Tense=fake|Voice=fake|Person=fake"
            "	uiduarum\r\n"
            ".	.	PUNC	MORPH=empty	.\r\n"
            ".	.	PUNC	MORPH=empty	.",
            result,
            "Punctuation should be reinserted and mostly should not break anything"
        )

    def test_leading_punctuation(self):
        """Check that consecutive punctation does not break anything

        Special case of consecutive dots, where sentences starts with it
        """
        tagger, data_iterator, formatter = make_controller([
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        result = tagger.tag_str(
            "( id enim ait ) turbabuntur a facie eius patris or phanorum et iudicis uiduarum .  .",
            formatter_class=formatter,
            iterator=data_iterator
        )
        self.assertIn(
            "form	lemma	POS	morph	treated_token\r\n"
            "(	(	PUNC	MORPH=empty	(\r\n"
            "id	id	fake	Case=fake|Numb=fake|Deg=fake|Mood=fake|Tense=fake|Voice=fake|Person=fake	id\r\n"
            "enim	enim	fake	Case=fake|Numb=fake|Deg=fake|Mood=fake|Tense=fake|Voice=fake|Person=fake	enim\r\n"
            "ait	ait	fake	Case=fake|Numb=fake|Deg=fake|Mood=fake|Tense=fake|Voice=fake|Person=fake	ait\r\n"
            ")	)	PUNC	MORPH=empty	)\r\n"
            "turbabuntur	turbabuntur	fake	Case=fake|Numb=fake|Deg=fake|Mood=fake|Tense=fake|Voice=fake|Person"
            "=fake	turbabuntur\r\n",
            result,
            "Leading punctuation should not break anything"
        )

    def test_punctuation_is_not_seen(self):
        """Check that punctuation is not seen by the tagger

        """
        tagger, data_iterator, formatter = make_controller([
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        tagger.tag_str(
            "( id enim ait ) turbabuntur a facie eius patris or phanorum et iudicis uiduarum .  .",
            formatter_class=formatter,
            iterator=data_iterator
        )
        self.assertNotIn(
            "(",
            [tok for sent in tagger.seen for tok in sent],
            "Punctuation should not be seen by the Tagger"
        )

    def test_j_are_temporarly_replaced(self):
        """Check that characters are replaced for the tagger, thus avoiding out of domain, and reinserted

        """
        tagger, data_iterator, formatter = make_controller([
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        result = tagger.tag_str(
            "( id enim ait ) turbabuntur a facie eius patris or phanorum et judicis uiduarum .  .",
            formatter_class=formatter,
            iterator=data_iterator
        )
        flatten_seen = list([tok for sent in tagger.seen for tok in sent])

        self.assertNotIn("judicis", flatten_seen, "'j' should be removed from tagging")
        self.assertIn("iudicis", flatten_seen, "And 'i' should replace it")
        self.assertIn("\njudicis\t", result, "But, in the end, the original form is given to the user")

    def test_underscores(self):
        string = "una operatio in ecclesiae fundamento.._... _ . laetatur autem pater quia filius perierat"
        tagger, data_iterator, formatter = make_controller([
            "una operatio in ecclesiae fundamento", "laetatur autem pater quia filius perierat"
        ])
        tagger.tag_str(
            string,
            formatter_class=formatter,
            iterator=data_iterator
        )
        flatten_seen = list([tok for sent in tagger.seen for tok in sent])
        self.assertEqual(
            ['una', 'operatio', 'in', 'ecclesiae', 'fundamento', 'laetatur', 'autem', 'pater', 'quia', 'filius',
             'perierat'],
            flatten_seen,
            "Seen element should not count the underscord"
        )


import regex as re
from ...pipeline.formatters.glue import GlueFormatter as SourceGlueFormatter

# Uppercase regexp
uppercase = re.compile("^[A-Z]$")


class GlueFormatter(SourceGlueFormatter):
    HEADERS = ["form", "lemma", "POS", "morph"]
    MORPH_PART = ["MODE", "TEMPS", "PERS.", "NOMB.", "GENRE", "CAS", "DEGRE"]

    PONCTU = re.compile(r"^\W+$")
    NUMBER = re.compile(r"\d+")
    PONFORT = [".", "...", "!", "?"]

    def __init__(self, tasks):
        self.tasks = tasks
        self.pos_tag = "POS"
        if "POS" not in self.tasks and "pos" in self.tasks:
            self.pos_tag = "pos"

    def rule_based(cls, token):
        if cls.PONCTU.match(token):
            lemma = token
            if token in GlueFormatter.PONFORT:
                pos = "PONfrt"
            else:
                pos = "PONfbl"
            return [token, lemma, pos, "MORPH=empty"]

    def format_line(self, token, tags, ignored=False):
        tags = list(tags)
        lemma = tags[self.tasks.index("lemma")]

        overwriten = self.rule_based(token)
        if overwriten:
            return overwriten

        if type(self).NUMBER.match(token):
            lemma = token
            tags[self.tasks.index(self.pos_tag)] = "ADJcar"

        return [
            token,
            lemma,
            tags[self.tasks.index(self.pos_tag)],
            "|".join(
                "{cat}={tag}".format(
                    cat=morph_part,
                    tag=tags[self.tasks.index(morph_part.replace(".", ""))]
                )
                for morph_part in GlueFormatter.MORPH_PART
                if morph_part.replace(".", "") in self.tasks and
                tags[self.tasks.index(morph_part.replace(".", ""))] != "_"
            ) or "MORPH=empty"
        ]



import os
from typing import Optional

from pie.tagger import Tagger
from pie import utils

from .pipeline.formatters.proto import Formatter
from .pipeline.disambiguators.proto import Disambiguator
from .pipeline.iterators.proto import DataIterator


class ExtensibleTagger(Tagger):
    def __init__(self, device='cpu', batch_size=100, lower=False, disambiguation=None):
        super(ExtensibleTagger, self).__init__(
            device=device,
            batch_size=batch_size,
            lower=lower
        )
        self.disambiguation: Optional[Disambiguator] = disambiguation

    def reinsert_full(self, formatter, sent_reinsertion, tasks):
        yield formatter.write_sentence_beginning()
        # If a sentence is empty, it's most likely because everything is in sent_reinsertions
        for reinsertion in sorted(list(sent_reinsertion.keys())):
            yield formatter.write_line(
                formatter.format_line(
                    token=sent_reinsertion[reinsertion],
                    tags=[""] * len(tasks)
                )
            )
        yield formatter.write_sentence_end()

    def tag_file(self, fpath: str, iterator: DataIterator, formatter_class: type):
        # Read content of the file
        with open(fpath) as f:
            data = f.read()

        _, ext = os.path.splitext(fpath)

        with open(utils.ensure_ext(fpath, ext, 'pie'), 'w+') as f:
            for line in self.iter_tag(data, iterator, formatter_class):
                f.write(line)

    def tag_str(self, data: str, iterator: DataIterator, formatter_class: type) -> str:
        return "".join(list(self.iter_tag(data, iterator, formatter_class)))

    def iter_tag(self, data: str, iterator: DataIterator, formatter_class: type):
        header = False
        formatter = None

        for chunk in utils.chunks(
                iterator(data, lower=self.lower),
                size=self.batch_size):
            # Unzip the batch into the sentences, their sizes and the dictionaries of things that needs
            #  to be reinserted
            sents, lengths, needs_reinsertion = zip(*chunk)
            # Removing punctuation might create empty sentences !
            #  Which would crash Torch
            empty_sents_indexes = {
                index: []
                for index, sent in enumerate(sents)
                if len(sent) == 0
            }
            tagged, tasks = self.tag(
                sents=[sent for sent in sents if len(sent)],
                lengths=lengths
            )
            formatter: Formatter = formatter_class(tasks)

            # We keep a real sentence index
            real_sentence_index = 0
            for sent in tagged:
                if not sent:
                    continue
                # Gets things that needs to be reinserted
                sent_reinsertion = needs_reinsertion[real_sentence_index]

                # If the header has not yet be written, write it
                if not header:
                    yield formatter.write_headers()
                    header = True

                # Some sentences can be empty and would have been removed from tagging
                #  we check and until we get to a non empty sentence
                #  we increment the real_sentence_index to keep in check with the reinsertion map
                while real_sentence_index in empty_sents_indexes:
                    yield from self.reinsert_full(
                            formatter,
                            needs_reinsertion[real_sentence_index],
                            tasks
                    )
                    real_sentence_index += 1

                yield formatter.write_sentence_beginning()

                # If we have a disambiguator, we run the results into it
                if self.disambiguation:
                    sent = self.disambiguation(sent, tasks)

                reinsertion_index = 0
                index = 0

                for index, (token, tags) in enumerate(sent):
                    while reinsertion_index + index in sent_reinsertion:
                        yield formatter.write_line(
                            formatter.format_line(
                                token=sent_reinsertion[reinsertion_index + index],
                                tags=[""] * len(tasks)
                            )
                        )
                        del sent_reinsertion[reinsertion_index + index]
                        reinsertion_index += 1

                    yield formatter.write_line(
                        formatter.format_line(token, tags)
                    )

                for reinsertion in sorted(list(sent_reinsertion.keys())):
                    yield formatter.write_line(
                        formatter.format_line(
                            token=sent_reinsertion[reinsertion],
                            tags=[""] * len(tasks)
                        )
                    )

                yield formatter.write_sentence_end()

                real_sentence_index += 1

            while real_sentence_index in empty_sents_indexes:
                yield from self.reinsert_full(
                    formatter,
                    needs_reinsertion[real_sentence_index],
                    tasks
                )
                real_sentence_index += 1

        if formatter:
            yield formatter.write_footer()



import sys
import regex as re
import click

try:
    import cltk
    from cltk.tokenize.word import WordTokenizer
except ImportError as E:
    click.echo(click.style("You need to install cltk and its Latin Data to runs this package", fg="red"))
    click.echo("pip install cltk")
    click.echo("pie-ext install-addons lasla")
    sys.exit(0)


from pie_extended.pipeline.iterators.proto import DataIterator
from pie_extended.pipeline.formatters.glue import GlueFormatter as SourceGlueFormatter
from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer as SourceMemorizingTokenizer


# Uppercase regexp
uppercase = re.compile(r"^[A-Z]$")


class MemorizingTokenizer(SourceMemorizingTokenizer):

    re_add_space_around_punct = re.compile(r"(\s*)([^\w\s\.])(\s*)")
    re_normalize_space = re.compile(r"(\s+)")
    re_sentence_tokenizer = re.compile(r"([_||[^\s\w]]+(?:[\s_||[\W]]+)?)", re.VERSION1)

    def __init__(self):
        self.tokens = [
        ]

        self._word_tokenizer = WordTokenizer("latin")

    def word_tokenizer(self, data):
        return self._word_tokenizer.tokenize(data)

    def sentence_tokenizer(self, data):
        sentences = list()
        first_is_dot = False
        started_writting = False  # Allows for avoiding to compute length
        for sent in MemorizingTokenizer.re_sentence_tokenizer.split(data):
            sent = sent.strip()
            if sent:
                if MemorizingTokenizer.re_sentence_tokenizer.match(sent):
                    if not started_writting:
                        sentences.append(sent)
                        first_is_dot = True
                    else:
                        sentences[-1] += " " + sent
                else:
                    if first_is_dot:
                        sentences[-1] += " " + sent
                        first_is_dot = False
                    else:
                        sentences.append(sent)

                if not started_writting and len(sentences):
                    started_writting = True

        yield from sentences

    def replacer(self, inp: str):
        inp = inp.replace("U", "V").replace("v", "u").replace("J", "I").replace("j", "i").lower()
        return inp

    def normalizer(self, data: str):
        # Fix regarding the current issue of apostrophe
        # https://github.com/cltk/cltk/issues/925#issuecomment-522065530
        # On the other hand, it creates empty tokens...
        data = MemorizingTokenizer.re_add_space_around_punct.sub(" \g<2> ", data)
        data = MemorizingTokenizer.re_normalize_space.sub(" ", data)
        return data


class GlueFormatter(SourceGlueFormatter):
    HEADERS = ["form", "lemma", "POS", "morph", "treated_token"]
    MORPH_PART = ["Case", "Numb", "Deg", "Mood", "Tense", "Voice", "Person"]
    PONCTU = re.compile(r"^\W+$")

    def __init__(self, tokenizer_memory):
        super(GlueFormatter, self).__init__([])
        self.tokenizer_memory = tokenizer_memory

    def rule_based(cls, token):
        if cls.PONCTU.match(token):
            return [token, token, "PUNC", "MORPH=empty", token]
        elif token.startswith("-"):
            if token == "-ne":
                lemma = "ne2"
            else:
                lemma = token[1:]
            return [token, lemma, "CONcoo", "MORPH=empty", token]

        return None


def get_iterator_and_formatter():
    tokenizer = MemorizingTokenizer()
    formatter = GlueFormatter(tokenizer)
    iterator = DataIterator(
        tokenizer=tokenizer,
        remove_from_input=DataIterator.remove_punctuation
    )
    return iterator, formatter



import os
from typing import Tuple, Iterable, Generator, Union
from importlib import import_module

import requests

from .. import models
from ..utils import Metadata, PATH, get_path
from ..tagger import ExtensibleTagger
from ..utils import ObjectCreator
from pie.utils import model_spec


def get_model(model):
    return import_module("{}.{}".format(models.__name__, model))


def download(module) -> Iterable[Union[str, int]]:
    lemmatizer = get_model(module)
    os.makedirs(os.path.join(PATH, module), exist_ok=True)
    yield len(lemmatizer.DOWNLOADS)
    for file in lemmatizer.DOWNLOADS:

        data = requests.get(file.url)
        new_path = get_path(module, file.name)

        with open(new_path, "wb") as f:
            f.write(data.content)
        yield file.name


def get_list() -> Iterable[Tuple[str, Metadata]]:
    for module in models.modules:
        desc = getattr(get_model(module), "DESC", None)
        if desc:
            yield module, desc


def get_tagger(model: str, batch_size: int = 16, device="cpu") -> ExtensibleTagger:
    module = get_model(model)
    disambiguator = getattr(module, "Disambiguator", None)
    if isinstance(disambiguator, ObjectCreator):
        disambiguator = disambiguator.create()
    tagger = ExtensibleTagger(disambiguation=disambiguator, batch_size=batch_size, device=device)
    for model, tasks in model_spec(getattr(module, "Models")):
        tagger.add_model(model, *tasks)
    return tagger


def tag_file(model: str, tagger: ExtensibleTagger, fpath):
    module = get_model(model)
    iterator, formatter = getattr(module, "get_iterator_and_formatter")()
    tagger.tag_file(fpath, iterator=iterator, formatter_class=formatter)
    return True


def get_addons(model: str):
    """ Runs the `addons` function from a module """
    module = get_model(model)
    addons = getattr(module, "addons", lambda : True)
    return addons()
